/* hgRefSeqStatus - Create the refSeqStatus table from loc2ref file. */
#include "common.h"
#include "linefile.h"
#include "hash.h"
#include "jksql.h"
/* auto-generated by autoSql: */
#include "refSeqStatus.h"

void usage()
/* Explain usage and exit. */
{
errAbort(
  "hgRefSeqStatus - Create the refSeqStatus table from a loc2ref file.\n"
  "usage:\n"
  "   hgRefSeqStatus database loc2ref\n"
  "where:\n"
  "   database - destination mysql database, often hgN for some N\n"
  "   loc2ref - wget ftp://ftp.ncbi.nih.gov/refseq/LocusLink/loc2ref\n"
  );
}

void readLoc2ref(char *fileName, struct hash **retStatusHash)
/* Read loc2ref file. Create hash of accession to status. */ 
{
struct lineFile *lf = lineFileOpen(fileName, TRUE);
char *line;
char *words[4];
int wordCount;
struct hash *statHash = newHash(0);
char *acc, *dot, *stat;
int lineCount = 0, count = 0;

while (lineFileNext(lf, &line, NULL))
    {
    ++lineCount;
    wordCount = chopTabs(line, words);
    if (wordCount == 4)
        {
	/* NOTE: we could look at all 6 fields... [5] is the taxonomy id,
                      human:  9606
                      mouse: 10090 
                        rat: 10116
                  zebrafish:  7955
           D. melanoagaster:  7227
	   But if we can trust the accession numbers to be different for 
	   different organisms (???) this will be OK.
	*/
	acc = cloneString(words[1]);
	/* trim the version number, if present */
	dot = strchr(acc, '.');
	if (dot != NULL)
	  *dot = 0;
	stat = cloneString(words[3]);
	stat[0] = toupper(stat[0]);
	hashAdd(statHash, acc, stat);
	count++;
	}
    }
lineFileClose(&lf);
*retStatusHash = statHash;
printf("Added %d acc statuses from %s in %d lines\n",
       count, fileName, lineCount);
}

void hgRefSeqStatus(char *database, char *loc2ref)
/* hgRefSeqStatus - Create the refSeqStatus table from a loc2ref file. */
{
struct hash *statusHash = NULL;	/* Hash of accession -> status. */
struct hashEl *el = NULL;
struct refSeqStatus rss;
struct sqlConnection *conn;
char *tabName = "refSeqStatus.tab";
FILE *f = NULL;
char query[256];

readLoc2ref(loc2ref, &statusHash);

conn = sqlConnect(database);

printf("Writing %s\n", tabName);
f = mustOpen(tabName, "w");
for (el = hashElListHash(statusHash);  el != NULL;  el = el->next)
    {
    /* Fill out a refSeqStatus data structure. */
    zeroBytes(&rss, sizeof(rss));
    rss.mrnaAcc = el->name;
    rss.status  = (char *)el->val;

    refSeqStatusTabOut(&rss, f);
    }
carefulClose(&f);

/* First make table definition. */
sqlRemakeTable(conn, "refSeqStatus",
	       /* is there a cleaner way to include the .sql?? */
	       "CREATE TABLE refSeqStatus (\n    mrnaAcc varchar(255) not null,      # RefSeq gene accession name\n    status varchar(255) not null,       # Status (Reviewed, Provisional, Predicted)\n              #Indices\n    PRIMARY KEY(mrnaAcc)\n)");

printf("Loading database %s\n", database);
sqlUpdate(conn, "delete from refSeqStatus");
sprintf(query, "load data local infile '%s' into table refSeqStatus", tabName);
sqlUpdate(conn, query);

sqlDisconnect(&conn);
}


int main(int argc, char *argv[])
/* Process command line. */
{
if (argc != 3)
    usage();
hgRefSeqStatus(argv[1], argv[2]);
return 0;
}
